const http = require('http'); // http模块
const url = require('url');   // 解析路径
const path = require('path'); // path模块
const fs = require('fs');     // 文件处理模块
const mime = require('mime'); // 解析文件类型
const crypto = require('crypto'); // 加密模块
const zlib = require('zlib');     // 压缩
const openbrowser = require('open'); //自动启动浏览器 
const handlebars = require('handlebars'); // 模版
const templates = require('./templates'); // 用来渲染的模版文件

class StaticServer {
  constructor(options) {
    this.host = options.host;
    this.port = options.port;
    this.rootPath = process.cwd();
    this.cors = options.cors;
    this.openbrowser = options.openbrowser;
  }
  requestHandler(req, res) {

  	// 通过url模块解析请求路径，获取请求文件
  	const {pathname} = url.parse(req.url);

  	// 获取请求的文件路径
  	const filepath = path.join(this.rootPath, pathname);

  	//判断文件是否存在 
  	fs.stat(filepath, (err, stat) => {
  		if(!err) {
  			if(stat.isDirectory()) {
  				this.responseDirectory(req, res, filepath, pathname);
  			}else {
  				this.responseFile(req, res, filepath, stat);
  			}
  		}else {
  			this.responseNotFound(req, res);
  		}
  	})
  }
  responseFile(req, res, filepath, stat) {
  	this.cacheHandle(req, res, filepath).then(data => {
  		if(data === true) {
  			res.writeHead(304);
  			res.end();
  		}else {
  			res.setHeader('Content-Type', mime.getType(filepath) + ';charset=utf-8');
  			res.setHeader('Etag', data);

  			this.cors && res.setHeader('Access-Control-Allow-Origin', '*');

  			const compress = this.compressHandler(req, res);

  			if(compress) {
  				fs.createReadStream(filepath).pipe(compress).pipe(res);
  			} else {
  				fs.createReadStream(filepath).pipe(res);
  			}
  		}
  	}, error => {
  		this.responseError(req, res, error);
  	})
  }
  responseDirectory(req, res, filepath, pathname) {
  	fs.readdir(filepath, (err, files) => {
  		if(!err) {
  			const fileList = files.map(file => {
  				const isDirectory = fs.statSync(filepath + '/' + file).isDirectory();
  				return {
  					filename: file,
  					url: path.join(pathname, file),
  					isDirectory
  				};
  			});
  			const html = handlebars.compile(templates.fileList)({
  				title: pathname,
  				fileList
  			});
  			res.setHeader('Content-Type', 'text/html');
  			res.end(html);
  		}
  	})
  }
  responseError(req, res, err) {
  	res,writeHead(500);
  	res.end(`there is something wrong in th server! please try later!`);
  }
  responseNotFound(req, res) {
  	const html = handlebars.compile(templates.notFound)();
  	res.writeHead(404, {

  		'Content-Type': 'text/html'
  	});
  	res.end(html);
  }
  cacheHandle(req, res, filepath) {
  	return new Promise((resolve, reject) => {
  		const readStream = fs.createReadStream(filepath);
  		const md5 = crypto.createHash('md5');
  		const ifNoneMatch = req.headers['if-none-match'];
  		readStream.on('data', data => {
  			md5.update(data);
  		});

  		readStream.on('end', () => {
  			let etag = md5.digest('hex');
  			if(ifNoneMatch === etag) {
  				resolve(true);
  			}
  			resolve(etag);
  		});

  		readStream.on('error', err => {
  			reject(err);
  		});
  	});
  }
  compressHandler(req, res) {
  	const acceptEncoding = req.headers['accept-encoding'];
  	if(/\bgzip\b/.test(acceptEncoding)) {
  		res.setHeader('Content-Encoding', 'gzip');
  		return zlib.createGzip();
  	}else if(/\bdeflate\b/.test(acceptEncoding)) {
  		res.setHeader('Content-Encoding', 'deflate');
  		return zlib.createDeflate();
  	}else{
  		return false;
  	}
  }
  start() {
  	const server = http.createServer((req, res) => this.requestHandler(req, res));
  	server.listen(this.port, () => {
  		if(this.openbrowser) {
  			openbrowser(`http://${this.host}:${this.port}`);
  		}
  		console.log(`server started in http://${this.host}:${this.port}`);
  	})
  }
}

module.exports = StaticServer;